# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import ruamel.yaml


def none_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:null', 'null')


class YAML(object):
    def __init__(self):
        self.yaml = ruamel.yaml.YAML(typ='rt')
        self.yaml.width = 256
        self.yaml.allow_duplicate_keys = True
        self.yaml.representer.add_representer(type(None), none_representer)
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.yaml.Constructor.add_constructor(
            '!encrypted/pkcs1-oaep',
            ruamel.yaml.SafeConstructor.construct_yaml_seq,
        )

    def load(self, stream):
        return self.yaml.load(stream)

    def dump(self, data, *args, **kwargs):
        self.yaml.dump(data, *args, **kwargs)



#!/usr/bin/env python3

# Copyright 2020 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys

import requests
import yaml

FILES_URL = "https://opendev.org/api/v1/repos/{}/git/trees/master"
IGNORED_REPOS = []

parser = argparse.ArgumentParser()
parser.add_argument(
    '-l', '--legacy-projects',
    default='./reference/legacy.yaml',
    help='legacy.yaml file path (%(default)s)'
)
args = parser.parse_args()

with open(args.legacy_projects, 'r', encoding='utf-8') as f:
    legacy_projects = yaml.safe_load(f.read())

errors = 0
for team_name, team_data in legacy_projects.items():
    # Check if the team has been retired
    if 'retired-on' in team_data:
        continue

    deliverables = team_data.get('deliverables')

    # Team has no deliverables, retired with no retired-on date
    if not deliverables:
        print('{} team has no deliverables with no retired-on date'.format(
            team_name
        ))
        errors += 1
        continue

    # In this case, team is not retired but has retired projects
    for deliverable_name, deliverable_data in deliverables.items():
        # Retired-on date missing for a deliverable
        if 'retired-on' not in deliverable_data:
            print('{} is missing a retired-on date'.format(deliverable_name))
            errors += 1
            continue

        # Ensure that the repositories has no content.
        for repo in deliverable_data['repos']:
            if repo in IGNORED_REPOS:
                msg = '{} is ignored'.format(repo)
                print(msg)
                continue

            if repo.startswith('opendev/'):
                msg = '{} not in openstack namespace, ignoring.'.format(repo)
                print(msg)
                continue

            url = FILES_URL.format(repo)
            files = requests.get(url).json()
            file_names = sorted([f['path'] for f in files['tree']])

            if file_names != ['.gitreview', 'README.rst']:
                msg = '{} is not properly retired, files: {}.'.format(
                    repo, file_names)
                print(msg)
                errors += 1
                continue


sys.exit(1 if errors else 0)



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import sys

import requests
import yaml

parser = argparse.ArgumentParser()
parser.add_argument(
    '-p', '--projects',
    default='./reference/projects.yaml',
    help='projects.yaml file path (%(default)s)',
)
parser.add_argument(
    '-l', '--legacy-projects',
    default='./reference/legacy.yaml',
    help='legact.yaml file path (%(default)s)'
)
parser.add_argument(
    '-g', '--gerrit',
    default=('http://opendev.org/openstack-infra/project-config/'
             'raw/branch/master/gerrit/projects.yaml'),
    help=('URL for gerrit project list, ignored if --project-config is set or '
          'when running in Zuul'),
)
parser.add_argument(
    '-c', '--project-config',
    default=('/home/zuul/src/opendev.org/openstack-infra/project-config'),
    help='Local path of project-config',
)
args = parser.parse_args()

with open(args.projects, 'r', encoding='utf-8') as f:
    projects = yaml.safe_load(f.read())

with open(args.legacy_projects, 'r', encoding='utf-8') as f:
    legacy_projects = yaml.safe_load(f.read())
legacy_repos = []
for team_name, team_data in legacy_projects.items():
    legacy_deliverables = team_data.get('deliverables')
    if not legacy_deliverables:
        continue
    for deliverable_name, deliverable_data in legacy_deliverables.items():
        for repo in deliverable_data['repos']:
            legacy_repos.append(repo)

if os.path.exists(args.project_config):
    projects_yaml = '%s/gerrit/projects.yaml' % args.project_config
    with open(projects_yaml) as gerrit_projects:
        gerrit_data = yaml.safe_load(gerrit_projects)
else:
    response = requests.get(args.gerrit)
    gerrit_data = yaml.safe_load(response.text)
gerrit_projects = set(
    entry.get('project')
    for entry in gerrit_data
)

errors = 0
for team_name, team_data in projects.items():
    deliverables = team_data.get('deliverables')
    for deliverable_name, deliverable_data in deliverables.items():
        for repo_name in deliverable_data.get('repos', []):
            if repo_name in legacy_repos:
                msg = ('Repository {} is present in project data {} as '
                       'well as in legecy project data {}')
                print(
                    msg.format(repo_name, args.projects, args.legacy_projects))
                errors += 1

            if repo_name not in gerrit_projects:
                print('Unknown repository {} as part of {} in {}'.format(
                    repo_name, deliverable_name, team_name))
                errors += 1

sys.exit(1 if errors else 0)



#!/usr/bin/env python
"""Extract relevant columns.
"""

import csv

with open('defcore.csv', 'r') as infile:
    reader = csv.reader(infile)
    read_iter = iter(reader)

    # Read past the 2 header lines
    read_iter.next()
    read_iter.next()

    rows = [[row[0]] + row[6:9] for row in read_iter]

    # Ignore the percentages at the bottom
    del rows[-1]

    # Add the spreadsheet row numbers
    for i, row in enumerate(rows, 3):
        row.insert(0, str(i))
    print(rows[0])
    rows[0][0] = 'Row'

    # Candidate Capabilities,TC Future Direction,Complete,Stable
    widths = []
    for i in range(len(rows[0])):
        widths.append(max(len(row[i])+2 for row in rows))

    fmt = '{0:>%d} {1:<%d} {2:^%d} {3:^%d} {4:^%d}' % tuple(widths)

    header = ' '.join('=' * wid for wid in widths)

    print(header)
    print(fmt.format(*rows[0]))
    print(header)
    for row in rows[1:]:
        print(fmt.format(*row))
    print(header)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslotest import base

from openstack_governance import _yamlutils
from openstack_governance import governance


_team_data_yaml = """
Release Management:
  ptl:
    name: Doug Hellmann
    irc: dhellmann
    email: doug@doughellmann.com
  irc-channel: openstack-release
  mission: >
    Coordinating the release of OpenStack deliverables, by defining the
    overall development cycle, release models, publication processes,
    versioning rules and tools, then enabling project teams to produce
    their own releases.
  url: https://wiki.openstack.org/wiki/Release_Management
  liaisons:
    release:
      - name: Hervé Beraud
        irc: hberaud
        email: hberaud@redhat.com
    tact-sig:
      - name: Hervé Beraud
        irc: hberaud
        email: hberaud@redhat.com
      - name: Daniel Bengtsson
        irc: damani
        email: dbengt@redhat.com
    security:
      - name: Daniel Bengtsson
        irc: damani
        email: dbengt@redhat.com
  deliverables:
    release-schedule-generator:
      repos:
        - openstack/release-schedule-generator
    release-test:
      repos:
        - openstack/release-test
    release-tools:
      repos:
        - openstack-infra/release-tools
    releases:
      repos:
        - openstack/releases
    reno:
      repos:
        - openstack/reno
      docs:
        contributor: https://docs.openstack.org/developer/reno/
    specs-cookiecutter:
      repos:
        - openstack-dev/specs-cookiecutter
"""

_tc_data_yaml = """
---
# List of repositories owned by the technical committee
Technical Committee:
  - repo: openstack/governance
  - repo: openstack/project-team-guide
"""

_sigs_data_yaml = """
---
# List of repositories owned by SIGs
meta:
  - repo: openstack/governance-sigs
api:
  - repo: openstack/api-sig
"""

TEAM_DATA = _yamlutils.loads(_team_data_yaml)
TC_DATA = _yamlutils.loads(_tc_data_yaml)
SIGS_DATA = _yamlutils.loads(_sigs_data_yaml)


class TestGetRepoOwner(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.gov = governance.Governance(TEAM_DATA, TC_DATA, SIGS_DATA)

    def test_repo_exists(self):
        owner = self.gov.get_repo_owner(
            'openstack/releases',
        )
        self.assertEqual('Release Management', owner)

    def test_repo_exists_tc(self):
        owner = self.gov.get_repo_owner(
            'openstack/governance',
        )
        self.assertEqual('Technical Committee', owner)

    def test_no_such_repo(self):
        self.assertRaises(
            ValueError,
            self.gov.get_repo_owner,
            'openstack/no-such-repo',
        )


class TestLocalRepo(base.BaseTestCase):

    def test_create(self):
        gov = governance.Governance.from_local_repo()
        repos = gov.get_repositories(
            team_name='Technical Committee',
        )
        repo_names = set(r.name for r in repos)
        self.assertIn('openstack/governance', repo_names)


class TestGetRepositories(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.gov = governance.Governance(TEAM_DATA, TC_DATA, SIGS_DATA)

    def test_by_team(self):
        repos = self.gov.get_repositories(
            team_name='Release Management',
        )
        self.assertEqual(
            sorted(['openstack/release-schedule-generator',
                    'openstack/release-test',
                    'openstack-infra/release-tools',
                    'openstack/releases',
                    'openstack/reno',
                    'openstack-dev/specs-cookiecutter']),
            sorted(r.name for r in repos),
        )

    def test_by_team_tc(self):
        repos = self.gov.get_repositories(
            team_name='Technical Committee',
        )
        self.assertEqual(
            sorted(['openstack/governance',
                    'openstack/project-team-guide']),
            sorted(r.name for r in repos),
        )

    def test_by_team_sig(self):
        repos = self.gov.get_repositories(
            team_name='meta SIG',
        )
        self.assertEqual(
            sorted(['openstack/governance-sigs']),
            sorted(r.name for r in repos),
        )

    def test_by_deliverable(self):
        repos = self.gov.get_repositories(
            deliverable_name='release-tools',
        )
        self.assertEqual(
            ['openstack-infra/release-tools'],
            [r.name for r in repos],
        )


class TestGetTeam(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.gov = governance.Governance(TEAM_DATA, TC_DATA, SIGS_DATA)

    def test_found(self):
        team = self.gov.get_team('meta SIG')
        self.assertEqual('meta SIG', team.name)

    def test_not_found(self):
        self.assertRaises(
            ValueError,
            self.gov.get_team,
            'No Such Team',
        )


class TestTeamProperties(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.team_with = governance.Team(
            'test',
            {'service': 'service name',
             'mission': 'mission statement'},
        )
        self.team_without = governance.Team(
            'test',
            {},
        )

    def test_service(self):
        self.assertEqual('service name', self.team_with.service)
        self.assertIsNone(self.team_without.service)

    def test_mission(self):
        self.assertEqual('mission statement', self.team_with.mission)
        self.assertIsNone(self.team_without.mission)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Show information about extra ATCs managed in this repo.
"""

import operator
import re

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import tables
from docutils.utils import SystemMessagePropagation
from sphinx.util import logging

import projects

LOG = logging.getLogger(__name__)


class ExtraATCsTable(tables.Table):
    """List the extra ATCs for the given project.
    """

    HEADERS = ('Full Name', 'Email', 'Expires In')
    HEADER_MAP = {
        'Full Name': 'name',
        'Email': 'email',
        'Expires In': 'expires-in',
    }

    option_spec = {'class': directives.class_option,
                   'name': directives.unchanged,
                   'project': directives.unchanged,
                   }

    def run(self):
        project = self.options.get('project')
        if not project:
            error = self.state_machine.reporter.error(
                'No project set for extra-atcs table',
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno)
            return [error]

        # Handle the width settings and title
        try:
            col_widths = self.get_column_widths(len(self.HEADERS))
            title, messages = self.make_title()
        except SystemMessagePropagation as detail:
            return [detail.args[0]]
        except Exception as err:
            error = self.state_machine.reporter.error(
                'Error processing memberstable directive:\n%s' % err,
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno,
            )
            return [error]

        all_teams = projects.get_project_data()
        try:
            team_data = all_teams[project]
        except KeyError:
            raise ValueError('No project %r in projects.yaml' % (project,))
        project_members = team_data.get('extra-atcs', [])

        # If we have no extra ATCs, skip building the table.
        if not project_members:
            return []

        # Sort the project_members by name
        project_members = sorted(
            project_members,
            key=operator.itemgetter('name'),
        )

        table_node = self.build_table(project_members, col_widths)
        table_node['classes'] += self.options.get('class', [])
        self.add_name(table_node)

        if title:
            table_node.insert(0, title)

        return [table_node] + messages

    def build_table(self, project_members, col_widths):
        table = nodes.table()

        # Set up the column specifications
        # based on the widths.
        tgroup = nodes.tgroup(cols=len(col_widths))
        table += tgroup
        tgroup.extend(nodes.colspec(colwidth=col_width)
                      for col_width in col_widths)

        # Set the headers
        thead = nodes.thead()
        tgroup += thead
        row_node = nodes.row()
        thead += row_node
        row_node.extend(
            nodes.entry(h, nodes.paragraph(text=h))
            for h in self.HEADERS
        )

        # The body of the table is made up of rows.
        # Each row contains a series of entries,
        # and each entry contains a paragraph of text.
        tbody = nodes.tbody()
        tgroup += tbody
        rows = []
        for row in project_members:
            trow = nodes.row()
            # Iterate over the headers in the same order every time.
            for h in self.HEADERS:
                # Get the cell value from the row data, replacing None
                # in re match group with empty string.
                cell = row.get(self.HEADER_MAP[h]) or ''
                entry = nodes.entry()
                para = nodes.paragraph(text=str(cell))
                entry += para
                trow += entry
            rows.append(trow)
        tbody.extend(rows)

        return table


_PATTERN = re.compile(r'(?P<project>.+):\s+(?P<name>.+)\s\((?P<email>.+)\)\s\[(?P<expires_in>.*)\]')


def setup(app):
    LOG.info('loading atcs extension')
    app.add_directive('extraatcstable', ExtraATCsTable)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Generate badges for the projects
"""

import os

from itertools import chain
from itertools import zip_longest
from PIL import ImageFont
from sphinx.util import logging

import projects

LOG = logging.getLogger(__name__)

NUM_COL = 4
PADDING = 8
BADGE_SPACING = 4
COLOR_SCHEME = {
    "brightgreen": "#4c1",
    "green": "#97CA00",
    "yellow": "#dfb317",
    "yellowgreen": "#a4a61d",
    "orange": "#fe7d37",
    "red": "#e05d44",
    "blue": "#007ec6",
    "grey": "#555",
    "lightgrey": "#9f9f9f",
}

OPENSTACK_SVG = """<svg id="Layer_1" x="0" y="0" height="20" width="20" data-name="Layer 1" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 209.67 180.35"><defs><style>.cls-1{opacity:0.98;}.cls-2{fill:#ed1944;}</style></defs><title>OpenStack_Logo_Mark</title><g class="cls-1"><path class="cls-2" d="M461.82,215.24h-150a17.17,17.17,0,0,0-17.12,17.12v40.35h41.61v-6.59a9.26,9.26,0,0,1,9.26-9.26h82.53a9.26,9.26,0,0,1,9.26,9.26v6.59H479V232.36A17.18,17.18,0,0,0,461.82,215.24Z" transform="translate(-294.67 -215.24)"/><path class="cls-2" d="M437.33,344.72a9.27,9.27,0,0,1-9.26,9.26H345.54a9.27,9.27,0,0,1-9.26-9.26v-6.59H294.67v40.34a17.17,17.17,0,0,0,17.12,17.13h150A17.18,17.18,0,0,0,479,378.47V338.13H437.33Z" transform="translate(-294.67 -215.24)"/><rect class="cls-2" y="69.37" width="41.62" height="41.62"/><rect class="cls-2" x="142.66" y="69.37" width="41.62" height="41.62"/></g><path class="cls-2" d="M504.33,386.39a9.2,9.2,0,1,0-9.2,9.21A9.21,9.21,0,0,0,504.33,386.39Zm-9.2,6.94a6.94,6.94,0,1,1,6.94-6.94A6.94,6.94,0,0,1,495.13,393.33Z" transform="translate(-294.67 -215.24)"/><path class="cls-2" d="M498.58,384.72v-.05a2.88,2.88,0,0,0-.76-2.09,3.38,3.38,0,0,0-2.45-.86H492v9h1.86v-3H495l1.66,3h2.14l-1.92-3.35A2.72,2.72,0,0,0,498.58,384.72Zm-1.88.06a1.3,1.3,0,0,1-1.47,1.35h-1.38v-2.72h1.34c1,0,1.51.45,1.51,1.35Z" transform="translate(-294.67 -215.24)"/></svg>"""

SVG_ROOT = """<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg contentScriptType="text/ecmascript" zoomAndPan="magnify"
contentStyleType="text/css" height="{height}" width="{width}"
preserveAspectRatio="xMidYMid meet" xmlns="http://www.w3.org/2000/svg"
 version="1.1" xmlns:xlink="http://www.w3.org/1999/xlink">
  <title id="os:gov:badges:title">
    This is a container for a set of OpenStack badges indicating the status and
    features of this project and its repository
  </title>
{svg}
</svg>
"""
FLAT_BADGE_TEMPLATE = """<svg id="{left_text}:{right_text}" width="{width}"
height="20" x="{svg_x}" y="{svg_y}">
<title>{left_text}:{right_text}</title>
<a target="_blank" xlink:href="{link}">
  <linearGradient id="smooth:{left_text}:{right_text}" x2="0" y2="100%">
    <stop offset="0" stop-color="#bbb" stop-opacity=".1"/>
    <stop offset="1" stop-opacity=".1"/>
  </linearGradient>

  <mask id="round:{left_text}:{right_text}">
    <rect width="{width}" height="20" rx="3" fill="#fff"/>
  </mask>

  <g mask="url('#round:{left_text}:{right_text}')">
    <rect width="{left_width}" height="20" fill="#555"/>
    <rect x="{left_width}" width="{right_width}" height="20" fill="{color}"/>
    <rect width="{width}" height="20"
          fill="url('#smooth:{left_text}:{right_text}')"/>
  </g>

  <g fill="#fff" text-anchor="middle"
     font-family="DejaVu Sans,Verdana,Geneva,sans-serif" font-size="11">
    <text x="{left_x}" y="15" fill="#010101"
          fill-opacity=".3">{left_text}</text>
    <text x="{left_x}" y="14">{left_text}</text>
    <text x="{right_x}" y="15" fill="#010101"
          fill-opacity=".3">{right_text}</text>
    <text x="{right_x}" y="14">{right_text}</text>
  </g>
</a>
</svg>
"""


def _badge(left_text, right_text, link=None, colorscheme='brightgreen'):

    font = ImageFont.truetype('DejaVuSans.ttf', 11)
    left_width = font.getsize(left_text)[0] + PADDING
    right_width = font.getsize(right_text)[0] + PADDING
    width = left_width + right_width

    data = {
        'link': link or '',
        'svg_x': 0,
        'svg_y': 0,
        'color': COLOR_SCHEME[colorscheme],
        'width': width,
        'left_width': left_width,
        'left_text': left_text,
        'left_x': left_width / 2,
        'right_width': right_width,
        'right_text': right_text,
        'right_x': left_width + right_width / 2 - 1,
    }

    return data


def _get_base_badges():
    return [
        _badge('openstack', 'community project',
               'https://governance.openstack.org/tc/reference/projects/'),
        _badge('cii best practices', 'passing',
               'https://bestpractices.coreinfrastructure.org/projects/246')
    ]


def _organize_badges(base_badges):

    # Arrange badges in NUM_COL columns, filling the rest with width=0 badges
    ziped = list(zip_longest(*(iter(base_badges),) * NUM_COL,
                             fillvalue={'width': 0}))

    result = []
    # Calculate x,y for each badge, leaving BADGE_SPACING between them
    for line, group in enumerate(ziped):
        # Start a new line at x=25, after the openstack logo
        result.append([])
        x = 25
        for col, badge in enumerate(group):
            # Skip width=0 badges
            if badge['width'] == 0:
                break

            # Column width is the width of the largest badge in column
            col_width = max(ziped, key=lambda s: s[col]['width'])[col]['width']

            badge['height'] = 20
            badge['svg_y'] = (20 + BADGE_SPACING) * line
            badge['svg_x'] = x
            x += col_width + BADGE_SPACING
            result[line].append(badge)
    return result


def _to_svg(badges):
    yield OPENSTACK_SVG
    for badge in badges:
        yield FLAT_BADGE_TEMPLATE.format(**badge)


def _generate_teams_badges(app, exception=None):
    LOG.info('Generating team badges')
    all_teams = projects.get_project_data()
    files = []

    badges_dir = os.path.join(app.outdir, 'badges')
    if not os.path.exists(badges_dir):
        os.mkdir(badges_dir)

    filename = os.path.join(badges_dir, 'project-unofficial.svg')
    svg_data = _badge('project', 'unofficial', colorscheme='red')
    svg = FLAT_BADGE_TEMPLATE.format(**svg_data)
    with open(filename, 'w', encoding='utf-8') as f:
        f.write(SVG_ROOT.format(height=20, width=106, svg=svg))
    files.append(filename)

    for team, info in all_teams.items():
        LOG.info('generating team badge for %s' % team)

        for name, deliverable in info['deliverables'].items():
            badges = _organize_badges(_get_base_badges())
            svg = '\n'.join(_to_svg(chain(*badges)))
            root_width = max([bdg_row[-1]['width'] + bdg_row[-1]['svg_x']
                              for bdg_row in badges])
            root_height = badges[-1][0]['svg_y'] + badges[-1][0]['height']

            for repo in deliverable.get('repos', []):
                repo_name = repo.split('/')[1]
                filename = os.path.join(badges_dir,
                                        '%s.svg' % projects.slugify(repo_name))
                with open(filename, 'w') as f:
                    f.write(SVG_ROOT.format(height=root_height,
                                            width=root_width, svg=svg))
                files.append(filename)

    return files


def setup(app):
    LOG.info('loading badges extension')
    app.connect('build-finished', _generate_teams_badges)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Do dirty things with wikis.
"""

import collections
import itertools

import mwclient


def get_page_section(page_content, section_start):
    "Return iterable of lines making up a section of a wiki page."
    lines = page_content.splitlines()
    lines = itertools.dropwhile(
        lambda x: x != section_start,
        lines,
    )
    next(lines)  # skip the start_section line
    lines = itertools.takewhile(
        lambda x: not x.startswith('='),  # another section or subsection
        lines,
    )
    return lines


def get_wiki_table(section_content):
    """Return iterable of dicts making up rows of a wiki table.

    Assumes there is only one table per section.

    """
    lines = itertools.dropwhile(
        lambda x: x != '{| class="wikitable"',
        section_content,
    )
    headings = []
    items = []
    for line in lines:
        if line == '|-':
            continue
        elif line.startswith('!'):
            headings = [h.strip() for h in line.lstrip('!').split('!!')]
        elif line in ['}', '|}']:
            # end of table
            break
        elif line.startswith('|'):
            items.extend(i.strip() for i in line.lstrip('|').split('||'))

        if len(items) == len(headings):
            row = {
                h: i
                for (h, i) in zip(headings, items)
            }
            yield row
            items = []


def get_wiki_page(name):
    "Return the text of a wiki page as a string."
    site = mwclient.Site('wiki.openstack.org')
    page = site.Pages[name]
    return page.text()


def get_liaison_data():
    "Return team -> liaisons mapping"
    project_to_liaisons = collections.OrderedDict()
    wiki_page = get_wiki_page('OpenStack_health_tracker')
    section = get_page_section(wiki_page, '=== Project Teams ===')
    table = get_wiki_table(section)

    for row in table:
        if not row:
            continue
        liaisons = [
            m.strip()
            for m in row['TC members'].split(',')
            if m.strip()
        ]
        project_to_liaisons[row['Group']] = liaisons

    return project_to_liaisons



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Verify that all data files match the schema."""

import argparse
import pkgutil

import jsonschema

from openstack_governance import yamltools


_yaml = yamltools.YAML()

_PROJECTS_SCHEMA = _yaml.load(
    pkgutil.get_data('openstack_governance',
                     'projects_schema.yaml').decode('utf-8')
)


def main():
    parser = argparse.ArgumentParser()
    parser.parse_args()

    errors = []

    with open('reference/projects.yaml', 'r', encoding='utf-8') as f:
        all_projects = _yaml.load(f.read())

    validator = jsonschema.Draft4Validator(_PROJECTS_SCHEMA)

    errors = False
    for e in validator.iter_errors(all_projects):
        errors = True
        print(e)

    return 1 if errors else 0



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Work with the governance repository."""

import logging
import os.path
import weakref

from openstack_governance import _yamlutils

import requests

LOG = logging.getLogger(__name__)
REPO_URL_BASE = "https://opendev.org/openstack/governance/raw/branch/master"


class Team(object):

    def __init__(self, name, data):
        self.name = name
        self.data = data
        # Protectively initialize the ptl data structure in case part
        # of it is missing from the project list, then replace any
        # values we do have from that data.
        self.ptl = {
            'name': 'MISSING',
            'irc': 'MISSING',
        }
        self.ptl.update(data.get('ptl', {}))
        self.deliverables = {
            dn: Deliverable(dn, di, self)
            for dn, di in self.data.get('deliverables', {}).items()
        }
        self.liaisons = data.get('liaisons', [])

    @property
    def service(self):
        return self.data.get('service')

    @property
    def mission(self):
        return self.data.get('mission')


class Deliverable(object):
    def __init__(self, name, data, team):
        self.name = name
        self.data = data
        self.team = weakref.proxy(team)
        self.repositories = {
            rn: Repository(rn, self)
            for rn in self.data.get('repos', [])
        }


class Repository(object):
    def __init__(self, name, deliverable):
        self.name = name
        self.deliverable = weakref.proxy(deliverable)


class Governance(object):

    _projects_filename = 'reference/projects.yaml'
    _tc_repos_filename = 'reference/technical-committee-repos.yaml'
    _sigs_repos_filename = 'reference/sigs-repos.yaml'

    def __init__(self, team_data, tc_data, sigs_data):
        self._team_data = team_data
        self._tc_data = tc_data
        self._sigs_data = sigs_data

        team_data['Technical Committee'] = {
            'deliverables': {
                repo['repo'].partition('/')[-1]: {'repos': [repo['repo']]}
                for repo in tc_data['Technical Committee']
            }
        }
        for sig_name, sig_info in sigs_data.items():
            team_data['{} SIG'.format(sig_name)] = {
                'deliverables': {
                    repo['repo'].partition('/')[-1]: {'repos': [repo['repo']]}
                    for repo in sig_info
                }
            }

        self._teams = [Team(n, i) for n, i in self._team_data.items()]

    @classmethod
    def from_local_repo(cls, repo_dir='.'):
        team_filename = os.path.join(repo_dir, cls._projects_filename)
        team_data = _yamlutils.load_from_file(team_filename)

        tc_filename = os.path.join(repo_dir, cls._tc_repos_filename)
        tc_data = _yamlutils.load_from_file(tc_filename)

        sigs_filename = os.path.join(repo_dir, cls._sigs_repos_filename)
        sigs_data = _yamlutils.load_from_file(sigs_filename)

        return cls(team_data, tc_data, sigs_data)

    @classmethod
    def from_remote_repo(cls, repo_url_base=REPO_URL_BASE, gittag=None):
        query_options = {}
        if gittag:
            query_options.update({'h': gittag})
        team_url = REPO_URL_BASE + '/reference/projects.yaml'
        LOG.debug('fetching team data from %s', team_url)
        r = requests.get(team_url, params=query_options)
        team_data = _yamlutils.loads(r.text)

        tc_url = REPO_URL_BASE + '/reference/technical-committee-repos.yaml'
        LOG.debug('fetching TC data from %s', tc_url)
        r = requests.get(tc_url, params=query_options)
        tc_data = _yamlutils.loads(r.text)

        sigs_url = REPO_URL_BASE + '/reference/sigs-repos.yaml'
        LOG.debug('fetching SIGs data from %s', sigs_url)
        r = requests.get(sigs_url, params=query_options)
        sigs_data = _yamlutils.loads(r.text)

        return cls(team_data, tc_data, sigs_data)

    def get_team(self, name):
        for team in self._teams:
            if team.name == name:
                return team
        raise ValueError('No team {!r} found'.format(name))

    def get_repo_owner(self, repo_name):
        """Return the name of the team that owns the repository.

        :param repo_name: Long name of the repository, such as 'openstack/nova'.

        """
        for team, info in self._team_data.items():
            for dname, dinfo in info.get('deliverables', {}).items():
                if repo_name in dinfo.get('repos', []):
                    return team
        raise ValueError('Repository %s not found in governance list' % repo_name)

    def get_repositories(self, team_name=None, deliverable_name=None):
        """Return a sequence of repositories, possibly filtered.

        :param team_name: The name of the team owning the repositories. Can be
            None. For Example: team_name='adjutant' or team_name='security SIG'
            or team_name='Technical Committee'
        :para deliverable_name: The name of the deliverable to which all
           repos should belong.

        """
        if team_name:
            teams = [self.get_team(team_name)]
        else:
            teams = self._teams

        for team in teams:
            if deliverable_name and deliverable_name not in team.deliverables:
                continue
            if deliverable_name:
                deliverables = [team.deliverables[deliverable_name]]
            else:
                deliverables = team.deliverables.values()
            for deliverable in deliverables:
                for repository in deliverable.repositories.values():
                    yield repository



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Build a table of the current members of the TC.
"""

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import tables
from docutils.utils import SystemMessagePropagation
from sphinx.util import logging

from openstack_governance import members

LOG = logging.getLogger(__name__)


class MembersTable(tables.Table):
    """Insert the members table using the referenced file as source.
    """

    HEADERS = ('Full Name', 'IRC', 'E-mail', 'Elected On', 'Role')
    HEADER_MAP = {
        'Full Name': 'name',
        'IRC': 'irc',
        'E-mail': 'email',
        'Elected On': 'date',
        'Role': 'role',
    }

    option_spec = {'class': directives.class_option,
                   'name': directives.unchanged,
                   'datafile': directives.unchanged,
                   }

    has_content = False

    def run(self):
        env = self.state.document.settings.env

        # The required argument to the directive is the name of the
        # file to parse.
        datafile = self.options.get('datafile')
        if not datafile:
            error = self.state_machine.reporter.error(
                'No filename in membertable directive',
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno)
            return [error]

        # Handle the width settings and title
        try:
            col_widths = self.get_column_widths(len(self.HEADERS))
            title, messages = self.make_title()
        except SystemMessagePropagation as detail:
            return [detail.args[0]]
        except Exception as err:
            error = self.state_machine.reporter.error(
                'Error processing memberstable directive:\n%s' % err,
                nodes.literal_block(self.block_text, self.block_text),
                line=self.lineno,
            )
            return [error]

        # Now find the real path to the file, relative to where we are.
        rel_filename, filename = env.relfn2path(datafile)

        # Build the table node using the parsed file
        data_iter = members.parse_members_file(filename)
        table_node = self.build_table(
            data_iter,
            col_widths,
        )
        table_node['classes'] += self.options.get('class', [])
        self.add_name(table_node)

        if title:
            table_node.insert(0, title)

        return [table_node] + messages

    def build_table(self, table_data, col_widths):
        table = nodes.table()

        # Set up the column specifications
        # based on the widths.
        tgroup = nodes.tgroup(cols=len(col_widths))
        table += tgroup
        tgroup.extend(nodes.colspec(colwidth=col_width)
                      for col_width in col_widths)

        # Set the headers
        thead = nodes.thead()
        tgroup += thead
        row_node = nodes.row()
        thead += row_node
        row_node.extend(
            nodes.entry(h, nodes.paragraph(text=h))
            for h in self.HEADERS
        )

        # The body of the table is made up of rows.
        # Each row contains a series of entries,
        # and each entry contains a paragraph of text.
        tbody = nodes.tbody()
        tgroup += tbody
        rows = []
        for row in table_data:
            trow = nodes.row()
            # Iterate over the headers in the same order every time.
            for h in self.HEADERS:
                # Get the cell value from the row data, replacing None
                # in re match group with empty string.
                cell = row.get(self.HEADER_MAP[h]) or ''
                entry = nodes.entry()
                para = nodes.paragraph(text=str(cell))
                entry += para
                trow += entry
            rows.append(trow)
        tbody.extend(rows)

        return table


def setup(app):
    LOG.info('loading members extension')
    app.add_directive('memberstable', MembersTable)



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import sys
from urllib import request
import yaml


TOP_X_OWNER = 10
# Zuul builds fetch size per query
BUILDSET_FETCH_SIZE = 500


class BaseQuery:
    def __init__(self, base_url, query_separator='&', verbose=False):
        self.verbose = verbose
        self.query_separator = query_separator
        if base_url.endswith('/'):
            self.base_url = base_url
        else:
            self.base_url = base_url + '/'

    def query(self, api, conditions=None):
        url = self.base_url + api
        if conditions:
            url += '?' + (self.query_separator.join(conditions))
        if self.verbose:
            print("Request on url %s" % url)
        result = request.urlopen(url)
        result = yaml.safe_load(result.read())
        if self.verbose:
            print(result)
        return result


class Zuul(BaseQuery):
    def get_project_buildsets(self, repos, fetch_size, branch='master', tenant='openstack'):
        # project={repos}&pipeline=gate&limit=500&branch=master

        conditions = ['pipeline=gate',
                      'limit='+str(fetch_size),
                      'branch=' + branch]
        for repo in repos:
            # count all repos in one set
            conditions.append('project=' + repo)
        return self.query(api='api/tenant/%s/buildsets' % tenant, conditions=conditions)

    def fetch_repo_buildset_success_rate(self, buildsets):
        success = {}
        not_success = {}
        for buildset in buildsets:
            if type(buildset) != dict:
                if self.verbose:
                    print("Buildset %s is not a dict. Skipping." % buildset)
                continue
            project = buildset['project']
            result = buildset['result']
            if result == 'SUCCESS':
                if project in success:
                    success[project] += 1
                else:
                    success[project] = 1
            else:
                if project in not_success:
                    not_success[project] += 1
                else:
                    not_success[project] = 1
        success_rates = {}
        all_projects = set(success).union(set(not_success))
        for project in all_projects:
            success_rates[project] = format(
                success.get(project, 0)/(
                    success.get(project, 0) + not_success.get(project, 0)
                )*100, '.0f') + '%'
        return success_rates


class Gerrit(BaseQuery):
    def query(self, api, conditions=None):
        url = self.base_url + api
        if conditions:
            # Set no-limit=1 to fetch larger amount of patches
            url += '?no-limit=1&q=' + (self.query_separator.join(conditions))
        if self.verbose:
            print("Request on url %s" % url)
        result = request.urlopen(url)
        if self.verbose:
            print(result)

        # The first line contains random chars like `'b\')]}\\\'`
        # Consider it as a bug and directly access second line from result
        result = yaml.safe_load(result.readlines()[1])
        if self.verbose:
            print(result)
        return result

    def get_project_query_list(self, repos):
        projects = []
        repoSize = len(repos)
        for i in range(repoSize):
            repo = repos[i]
            if i == 0:
                projects += ['(project:' + repo, 'OR']
            elif i == repoSize - 1:
                projects.append('project:' + repo + ')')
            else:
                projects += ['project:' + repo, 'OR']
        return projects

    def get_total_valid(self, repos, within, branch='master'):
        # q=(status:open+OR+status:merged)+project:{repo}+
        # NOT+label:Workflow<=-1+label:Verified>=1+NOT+age:{within}day

        conditions = ['(status:open', 'OR', 'status:merged)',
                      'NOT', 'label:Workflow<=-1',
                      'label:Verified>=1',
                      'branch:' + branch,
                      'NOT', 'age:' + str(within) + 'day']
        conditions += self.get_project_query_list(repos)
        return self.query(api='changes/', conditions=conditions)

    def get_total_not_review(self, repos, within, branch='master'):
        # q=status:open+project:{repos}+label:Code-Review=0+
        # NOT+label:Workflow<=-1+label:Verified>=1+NOT+age:{within}day

        conditions = ['status:open',
                      'NOT', 'label:Workflow<=-1',
                      'label:Verified>=1',
                      'label:Code-Review=0',
                      'branch:' + branch,
                      'NOT', 'age:' + str(within) + 'day']
        conditions += self.get_project_query_list(repos)
        return self.query(api='changes/', conditions=conditions)

    def get_total_merged(self, repos, within, branch='master'):
        # q=status:merged+project:{repos}+NOT+age:{within}day

        conditions = ['status:merged',
                      'branch:' + branch,
                      'NOT', 'age:' + str(within) + 'day']
        conditions += self.get_project_query_list(repos)
        return self.query(api='changes/', conditions=conditions)

    def count_owners(self, patches, limit=-1):
        owners = {}
        patchCount = len(patches)
        for p in patches:
            owner = p.get('owner', {}).get('_account_id', 0)
            if owner in owners:
                owners[owner] += 1
            else:
                owners[owner] = 1
        sorted_owners = sorted(
            owners.items(), reverse=True, key=lambda item: item[1])
        results = {}
        for owner, count in sorted_owners:
            owner = self.get_account_name(owner)
            if limit != -1 and len(results) >= limit:
                break
            results[owner] = format(
                (count / patchCount) * 100, '.2f') + '%'
        return results

    def get_account_name(self, account_id):
        # Profile will looks like: {"_account_id":22816,
        # "name":"OpenStack Release Bot",
        # "email":"infra-root@openstack.org","username":"release"}
        profile = self.query(api=('accounts/%s' % account_id))
        return profile.get('name', account_id)


def validate_gerrit_data(repos, count_days, branch, verbose):
    gerrit = Gerrit(
        base_url='https://review.opendev.org/', query_separator='+',
        verbose=verbose)
    print(" Validating Gerrit...")
    total_valid_patches = gerrit.get_total_valid(
        repos, within=count_days, branch=branch)

    if not total_valid_patches:
        print(" * There are no any patches proposed within "
              "%s days" % count_days)
        return

    not_reviewed_patches = gerrit.get_total_not_review(
        repos, within=count_days, branch=branch)
    merged_patches = gerrit.get_total_merged(
        repos, within=count_days, branch=branch)
    unreview_rate = (
        len(not_reviewed_patches)/len(total_valid_patches))*100//1
    merged_rate = (
        len(merged_patches)/len(total_valid_patches))*100//1
    owners_rates = gerrit.count_owners(
        patches=total_valid_patches, limit=10)

    print(" *", "There are",
          "%s ready for" % len(total_valid_patches),
          "review patches generated within %s days" % count_days)
    print(" *", "There are",
          "%s not reviewed" % len(not_reviewed_patches),
          "patches generated within %s days" % count_days)
    print(" *", "There are %s merged" % len(merged_patches),
          "patches generated within %s days" % count_days)
    print(" *", "Unreviewed patch rate for",
          "patches generated within",
          "%s days is %s" % (count_days, unreview_rate), '%')
    print(" *", "Merged patch rate for patches generated within",
          "%s days is %s" % (count_days, merged_rate), '%')
    print(" * ", "Here's top %s owner for patches" % TOP_X_OWNER,
          "generated within",
          "%s days (Name/Account_ID: Percentage):" % count_days)
    for owner, rate in owners_rates.items():
        print('    - ', owner, ': ', rate)


def validate_zuul_data(repos, branch, verbose):
    zuul = Zuul(base_url="https://zuul.opendev.org/", verbose=verbose)
    print(" Validate Zuul...")
    print(" Set buildsets fetch size to",
          "%s" % BUILDSET_FETCH_SIZE)
    gate_builds = zuul.get_project_buildsets(
        repos, fetch_size=BUILDSET_FETCH_SIZE, branch=branch)
    repo_success_rates = zuul.fetch_repo_buildset_success_rate(
        gate_builds)
    for repo, rate in repo_success_rates.items():
        print(" * Repo: %s gate job builds success" % repo,
              "rate: %s" % rate)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--file',
        default='./reference/projects.yaml',
        help='projects.yaml file path (%(default)s)',
    )
    parser.add_argument(
        '-p', '--projects',
        help='projects to analyze. Separate with comma',
    )
    parser.add_argument(
        '-d', '--days',
        default='180',
        help='Days to count gerrit patches generated since (%(default)s)',
    )
    parser.add_argument(
        '-b', '--branch',
        default='master',
        help='Branch to analyze on projects (%(default)s)',
    )
    parser.add_argument(
        '-z', '--skip-zuul',
        action='store_true',
        help='Skip Zuul analysis (%(default)s)',
    )
    parser.add_argument(
        '-g', '--skip-gerrit',
        action='store_true',
        help='Skip Gerrit analysis (%(default)s)',
    )
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Show debug information',
    )
    args = parser.parse_args()
    with open(args.file, 'r', encoding='utf-8') as f:
        projects = yaml.safe_load(f.read())

    project_scope = args.projects.split(',') if (
        args.projects is not None) else list(projects.keys())
    count_days = args.days

    def get_repos(project):
        '''Get repositories in project'''
        repos = []
        defn = projects.get(project, {})
        deliverables = defn.get('deliverables', {})
        for key, deliverable in deliverables.items():
            repos += deliverable.get('repos', '')
        return repos

    all_branch = args.branch.split(',')
    for branch in all_branch:
        if not (branch.startswith('master') or branch.startswith('stable/')):
            print('Invalid branch %s' % branch)
            sys.exit(1)
    for project in project_scope:
        print('*' * 50)
        if project not in projects:
            print('Invalid project %s' % project)
            continue
        print("Start Project %s analysis..." % project)
        repos = get_repos(project)
        print(" Includes repositories: %s" % repos)
        for branch in all_branch:
            print(" Start analysis branch %s..." % branch)
            if not args.skip_gerrit:
                validate_gerrit_data(repos, count_days, branch, args.verbose)

            if not args.skip_zuul:
                validate_zuul_data(repos, branch, args.verbose)

        print('*'*50)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse

import yaml


parser = argparse.ArgumentParser()
parser.add_argument(
    '-p', '--projects',
    default='./reference/projects.yaml',
    help='projects.yaml file path (%(default)s)',
)
format_grp = parser.add_mutually_exclusive_group()
format_grp.add_argument(
    '-f', '--full',
    action='store_const',
    const='full',
    dest='mode',
    default='full',
    help='show all details (default)',
)
format_grp.add_argument(
    '-e', '--email',
    action='store_const',
    const='email',
    dest='mode',
    help='show only email addresses',
)
format_grp.add_argument(
    '-i', '--irc',
    action='store_const',
    const='irc',
    dest='mode',
    help='show only IRC nics',
)
format_grp.add_argument(
    '-c', '--csv',
    action='store_const',
    const='csv',
    dest='mode',
    help='all data, comma separated',
)
args = parser.parse_args()


with open(args.projects, 'r', encoding='utf-8') as f:
    projects = yaml.safe_load(f.read())

fmt_str = {
    'full': '{name}\t{irc}\t{email}\t{project_name}',
    'irc': '{irc}',
    'email': '{email}',
    'csv': '"{name}","{irc}","{email}","{project_name}"',
}[args.mode]

for project_name, project_data in projects.items():
    ptl = project_data['ptl']
    print(fmt_str.format(project_name=project_name, **ptl))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Report on the current list of projects.
"""

from docutils import nodes
from docutils.parsers import rst
from docutils import statemachine
from sphinx.util import logging
from sphinx.util.nodes import nested_parse_with_titles

import projects

LOG = logging.getLogger(__name__)

IRC_LOG_URL_BASE = 'http://eavesdrop.openstack.org/irclogs/%23'


def _team_to_rst(name, info):

    if 'service' in info:
        title = "{0} ({1})".format(name.title(), info['service'])
    elif name == 'I18n':
        title = name
    else:
        title = name.title()

    yield '.. _project-%s:' % projects.slugify(name)
    yield ''
    yield '=' * len(title)
    yield title
    yield '=' * len(title)
    yield ''
    yield ':Home Page: ' + info.get('url', '')
    ptl = info.get('ptl', {'name': '', 'irc': '', 'email': ''})
    leadership_type = info.get('leadership_type')
    if leadership_type:
        yield ':Leadership Type: ' + leadership_type
    else:
        yield ':PTL: %(name)s (``%(irc)s``) <%(email)s>' % ptl
    irc_channel = info.get('irc-channel')
    if irc_channel:
        yield ':IRC Channel: `#%s <%s%s>`__' % (
            irc_channel, IRC_LOG_URL_BASE, irc_channel)
    service = info.get('service')
    if service:
        yield ':Service: ' + service
    liaisons = info.get('liaisons')
    if liaisons:
        contact_format = {'name': '', 'irc': '', 'email': ''}
        release = liaisons.get('release', contact_format)
        if release != contact_format:
            yield ':Release Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % rl
                for rl in release)
        tact_sig = liaisons.get('tact-sig', contact_format)
        if tact_sig != contact_format:
            yield ':TACT SIG Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % tl
                for tl in tact_sig)
        security = liaisons.get('security', contact_format)
        if security != contact_format:
            yield ':Security Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % sl
                for sl in security)
        events = liaisons.get('events', contact_format)
        if events != contact_format:
            yield ':Events Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % el
                for el in events)
        project_update_onboarding = liaisons.get('project_update_onboarding', contact_format)
        if project_update_onboarding != contact_format:
            yield ':Project Update Onboarding Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % pl
                for pl in project_update_onboarding)
        meeting_facilitator = liaisons.get('meeting_facilitator', contact_format)
        if meeting_facilitator != contact_format:
            yield ':Meeting Facilitator Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % ml
                for ml in meeting_facilitator)
        bug_deputy = liaisons.get('bug_deputy', contact_format)
        if bug_deputy != contact_format:
            yield ':Bug Deputy Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % bl
                for bl in bug_deputy)
        rfe_coordinator = liaisons.get('rfe_coordinator', contact_format)
        if rfe_coordinator != contact_format:
            yield ':RFE Coordinator Liaisons: ' + ', '.join(
                '%(name)s (``%(irc)s``) <%(email)s>' % rcl
                for rcl in rfe_coordinator)
    yield ''
    mission = info.get('mission', '').rstrip()
    if mission:
        yield "Mission"
        yield '-------'
        yield ''
        yield mission
        yield ''
    yield 'Deliverables'
    yield '------------'
    yield ''
    deliverables = info.get('deliverables', [])
    if deliverables:
        for repo_name, deliverable in deliverables.items():
            yield repo_name
            yield '~' * len(repo_name)
            yield ''
            yield ':Repositories: ' + ', '.join(
                ':repo:`%s`' % repo
                for repo in deliverable.get('repos', [])
            )
            yield ''
    else:
        yield 'None'
    yield ''
    if info.get('extra-atcs', []):
        yield 'Extra ATCs'
        yield '-----------'
        yield '.. extraatcstable::'
        yield '   :project: %s' % name
        yield ''


def _write_team_pages():
    all_teams = projects.get_project_data()
    files = []
    for team, info in all_teams.items():
        slug = projects.slugify(team)
        filename = 'reference/projects/%s.rst' % slug
        LOG.info('generating team page for %s' % team)
        with open(filename, 'w', encoding='utf-8') as f:
            f.write('\n'.join(_team_to_rst(team, info)))
        files.append(filename)
    return files


class TeamsListDirective(rst.Directive):

    has_content = False

    def run(self):
        all_teams = projects.get_project_data()

        # Build the view of the data to be parsed for rendering.
        result = statemachine.ViewList()
        for team_name in sorted(all_teams.keys()):
            team_info = all_teams[team_name]
            for line in _team_to_rst(team_name, team_info):
                result.append(line, '<' + __name__ + '>')

        # Parse what we have into a new section.
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)

        return node.children


def setup(app):
    LOG.info('loading teams extension')
    app.add_directive('teamslist', TeamsListDirective)
    _write_team_pages()



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from datetime import datetime
import json
import logging

import requests
import yaml

LOG = logging.getLogger()


# The OpenStack foundation member directory lookup API endpoint
MEMBER_LOOKUP_URL = 'https://openstackid-resources.openstack.org/'


def lookup_member(email):
    "A requests wrapper to querying the OSF member directory API"
    # URL pattern for querying foundation members by E-mail address
    LOG.debug('looking up %s', email)
    raw = requester(
        MEMBER_LOOKUP_URL + '/api/public/v1/members',
        params={
            'filter[]': [
                'group_slug==foundation-members',
                'email==' + email,
            ],
            'expand': 'all_affiliations',
        },
        headers={'Accept': 'application/json'},
    )
    decoded = decode_json(raw)
    try:
        return decoded['data'][0]
    except (KeyError, IndexError):
        return None


def requester(url, params={}, headers={}):
    """A requests wrapper to consistently retry HTTPS queries

    :param url: The URL to get.
    :type url: str
    :param params: Additional parameters to provide.
    :type params: dict(str, str)
    :param headers: Additional headers to set.
    :type params: dict(str, str)

    """
    # Try up to 3 times
    retry = requests.Session()
    retry.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
    return retry.get(url=url, params=params, headers=headers)


def decode_json(raw):
    """Trap JSON decoding failures and provide more detailed errors

    Remove ')]}' XSS prefix from data if it is present, then decode it
    as JSON and return the results.

    :param raw: Response text from API
    :type raw: str

    """

    # Gerrit's REST API prepends a JSON-breaker to avoid XSS vulnerabilities
    if raw.text.startswith(")]}'"):
        trimmed = raw.text[4:]
    else:
        trimmed = raw.text

    # Try to decode and bail with much detail if it fails
    try:
        decoded = json.loads(trimmed)
    except Exception:
        LOG.error(
            '\nrequest returned %s error to query:\n\n    %s\n'
            '\nwith detail:\n\n    %s\n',
            raw, raw.url, trimmed)
        raise
    return decoded


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-p', '--projects',
        default='./reference/projects.yaml',
        help='projects.yaml file path (%(default)s)',
    )
    args = parser.parse_args()

    # Quiet the urllib3 module output coming out of requests.
    logging.getLogger('urllib3').setLevel(logging.WARNING)

    logging.basicConfig()

    with open(args.projects, 'r', encoding='utf-8') as f:
        projects = yaml.safe_load(f.read())

    exit_code = 0

    for project_name, project_data in sorted(projects.items()):
        atcs = project_data.get('extra-atcs', [])
        for atc in atcs:
            try:
                member = lookup_member(atc['email'])
            except Exception as e:
                print('ERROR: {}: Could not validate ATC {}: {}'.format(
                    project_name, atc, e))
            else:
                if not member:
                    # NOTE(mnaser): The ATC membership checks were not
                    #               enforced for all ATCs expiring before
                    #               January 2020.  This should be removed
                    #               after January 2020.
                    expires = datetime.strptime(atc['expires-in'], "%B %Y")
                    if expires <= datetime(2020, 1, 1):
                        msg = 'Skipping {} from validation'.format(atc)
                    else:
                        msg = 'Unable to find membership: {}'.format(atc)
                        exit_code = 1

                    print('{}: {}'.format(project_name, msg))

    return exit_code


if __name__ == '__main__':
    import sys
    sys.exit(main())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Load the projects.yaml file.
"""

import copy
import os.path

from openstack_governance import projects
from sphinx.util import logging

LOG = logging.getLogger(__name__)

_projects_yaml = {}


def get_project_data():
    """Return a copy of the project data."""
    return copy.deepcopy(_projects_yaml)


def slugify(name):
    """Convert name to slug form for references."""
    return name.lower().replace(' ', '-')


def setup(app):
    global _projects_yaml

    filename = os.path.abspath('reference/projects.yaml')
    LOG.info('reading %s' % filename)
    _projects_yaml = projects.load_project_file(filename)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml
import yamlordereddictloader


def loads(blob):
    """Load a yaml blob and retain key ordering."""
    # This does use load, which is unsafe, but should be ok
    # for what we are loading here in this program; we should
    # be able to fix that in the future (if it matters).
    return yaml.load(blob, Loader=yamlordereddictloader.Loader)


def load_from_file(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        return loads(f.read())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import os.path
import subprocess

from sphinx.util import logging

LOG = logging.getLogger('page_context')

_projects_last_updated = datetime.datetime.now()
_projects_last_updated_set = False


def _get_last_updated_file(src_file):
    if not os.path.exists(src_file):
        return None
    try:
        last_updated_t = subprocess.check_output(
            [
                'git', 'log', '-n1', '--format=%ad',
                '--date=format:%Y-%m-%d %H:%M:%S',
                '--', src_file,
            ]
        ).decode('utf-8').strip()
    except subprocess.CalledProcessError as err:
        LOG.info('[governance] Could not get modification time of %s: %s',
                 src_file, err)
    else:
        if last_updated_t:
            try:
                return datetime.datetime.strptime(last_updated_t,
                                                  '%Y-%m-%d %H:%M:%S')
            except ValueError:
                LOG.info('[governance] Could not parse modification time of '
                         '%s: %r',
                         src_file, last_updated_t)
    return None


def html_page_context(app, pagename, templatename, context, doctree):
    # Use the last modified date from git instead of applying a single
    # value to the entire site.
    global _projects_last_updated
    global _projects_last_updated_set

    # Note: openstackdocstheme now handles this, so we only need to do our own
    # evaluation for generated reference pages.
    if pagename.startswith('reference/projects/'):
        if not _projects_last_updated_set:
            last_updated = _get_last_updated_file('reference/projects.yaml')
            if last_updated:
                LOG.info('[governance] Last updated for reference/'
                         'projects.yaml is %s', last_updated)
                _projects_last_updated = last_updated
                _projects_last_updated_set = True

        context['last_updated'] = _projects_last_updated


def setup(app):
    LOG.info('[governance] connecting html-page-context event handler')
    app.connect('html-page-context', html_page_context)
    return {
        'parallel_read_safe': True,
    }



#!/usr/bin/env python3

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import collections
import datetime
import json
import logging
import math
import operator

import prettytable
import requests

from openstack_governance import governance
from openstack_governance import members


LOG = logging.getLogger(__name__)

TC_SIZE = 9

# For formal-vote votes, age in days that a change has to have existed
# before it can be approved
creation_age_threshold = datetime.timedelta(days=7)


def decode_json(raw):
    """Trap JSON decoding failures and provide more detailed errors

    Remove ')]}' XSS prefix from data if it is present, then decode it
    as JSON and return the results.

    :param raw: Response text from API
    :type raw: str

    """

    # Gerrit's REST API prepends a JSON-breaker to avoid XSS vulnerabilities
    if raw.text.startswith(")]}'"):
        trimmed = raw.text[4:]
    else:
        trimmed = raw.text

    # Try to decode and bail with much detail if it fails
    try:
        decoded = json.loads(trimmed)
    except Exception:
        LOG.error(
            '\nrequest returned %s error to query:\n\n    %s\n'
            '\nwith detail:\n\n    %s\n',
            raw, raw.url, trimmed)
        raise
    return decoded


def query_gerrit(offset=0):
    """Query the Gerrit REST API"""
    url = 'https://review.opendev.org/changes/'
    LOG.debug('fetching %s', url)
    raw = requests.get(
        url,
        params={
            'n': '100',
            'start': offset,
            'q': 'project:openstack/governance is:open',
            'o': [
                'ALL_REVISIONS',
                'REVIEWER_UPDATES',
                'DETAILED_ACCOUNTS',
                'CURRENT_COMMIT',
                'LABELS',
                'DETAILED_LABELS',
                'MESSAGES',
            ],
        },
        headers={'Accept': 'application/json'},
    )
    return decode_json(raw)


def to_datetime(s, default=None):
    "Convert a string to a datetime.datetime instance"
    # Ignore the trailing decimal seconds.
    if s is None:
        return default
    s = s.rpartition('.')[0]
    return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')


def find_latest_revision(review):
    result = None
    result_date = None
    for rev in review.get('revisions', {}).values():
        rev_date = to_datetime(rev.get('created'))
        if not result or rev_date > result_date:
            result = rev
            result_date = rev_date
    return result


def count_votes(review, group='Rollcall-Vote'):
    votes = collections.Counter()
    votes.update(
        vote.get('value')
        for vote in review['labels'].get(group, {}).get('all', [])
    )
    if None in votes:
        del votes[None]
    return votes


def when_majority(review, required_count):
    "Return the date when the vote reached the required count."
    n = 0
    votes = review['labels'].get('Rollcall-Vote', {}).get('all', [])
    now = datetime.datetime.now()
    for vote in sorted(votes, key=lambda x: to_datetime(x.get('date'), now)):
        if vote.get('value'):
            n += 1
        if n >= required_count:
            d = vote.get('date')
            if d:
                return to_datetime(d)


def format_votes(votes):
    return 'nay:{:2d} / abs:{:2d} / yes:{:2d}'.format(
        votes.get(-1, 0), votes.get(0, 0), votes.get(1, 0)
    )


def get_votes_by_person(member, review):
    for label in ['Code-Review', 'Rollcall-Vote']:
        for vote in review['labels'].get(label, {}).get('all', []):
            if member['gerritid'] == vote['_account_id']:
                yield vote


def has_approved(member, review):
    return any(
        vote.get('value', 0) == 1
        for vote in get_votes_by_person(member, review)
    )


def has_rejected(member, review):
    return any(
        vote.get('value', 0) == -1
        for vote in get_votes_by_person(member, review)
    )


def has_commented(member, review):
    desired_revision = max(
        r.get('_number', -1)
        for r in review.get('revisions', {}).values()
    )
    for msg in review.get('messages', []):
        if msg.get('_revision_number', -1) != desired_revision:
            continue
        if msg.get('author', {}).get('_account_id', '') == member['gerritid']:
            return True


def all_changes():
    offset = 0
    while True:
        changes = query_gerrit(offset)

        yield from changes

        if changes and changes[-1].get('_more_changes', False):
            offset += 100
        else:
            break


def get_one_status(change, delegates, tc_members):
    topic = change.get('topic', 'unknown topic')
    subject = change.get('subject')
    owner = change.get('owner', {}).get('name')
    url = 'https://review.opendev.org/{}\n'.format(change['_number'])

    latest = find_latest_revision(change)
    latest_created = to_datetime(latest['created'])
    now = datetime.datetime.utcnow()
    age = now - latest_created
    earliest = ''

    code_reviews = count_votes(change, 'Code-Review')
    votes = count_votes(change)
    workflow = count_votes(change, 'Workflow')
    verified = count_votes(change, 'Verified')

    can_approve = 'no'

    if workflow[-1]:
        can_approve = 'WIP'

    elif workflow[1]:
        can_approve = 'APPROVED'

    elif verified[-1]:
        can_approve = 'NO, verification failure'
        earliest = 'when passing'

    elif topic == 'on-hold':
        can_approve = 'on hold'

    elif topic == 'formal-vote':
        # https://governance.openstack.org/tc/reference/charter.html#motions
        parts = []

        # At least at third rounded up positive votes and more positive than
        # negative votes.
        necessary_votes = math.ceil(TC_SIZE / 3)
        votes_to_approve = votes[1] >= necessary_votes and votes[1] > votes[-1]
        reached_necessary_votes = when_majority(change, necessary_votes)

        parts.append('last change on {}'.format(
            latest_created.isoformat(timespec='minutes')
        ))

        # At least creation_age_threshold days old.
        if age < creation_age_threshold:
            time_to_approve = False
            parts.append("has not been open %d days" %
                         creation_age_threshold.days)
            earliest = str(latest_created + creation_age_threshold)
        elif reached_necessary_votes:
            time_to_approve = True
        else:
            time_to_approve = False
            earliest = "after {} positive votes".format(necessary_votes)

        if votes_to_approve and time_to_approve:
            parts.append('YES')

        if reached_necessary_votes and not time_to_approve:
            parts.append('enough required votes but too soon')

        # Even if we can approve it, if there are dissenting votes we
        # may want to continue discussion or refine the proposal.
        if votes[-1]:
            if votes[-1] >= votes[1]:
                parts.append('too many dissenting votes')
            else:
                parts.append('dissenting votes')

        if votes[1] < necessary_votes:
            parts.append('not enough votes')
        elif votes[1] == necessary_votes:
            parts.append('minimum favorable votes')
        else:
            parts.append('sufficient votes')

        if not (votes[1] > math.floor(TC_SIZE / 2)):
            # Even if we can approve it, if the majority have not
            # voted yes we may want to continue discussion or call for
            # a final vote.
            parts.append('plz whip votes - no majority and things can change')

        can_approve = ',\n'.join(parts)

    elif topic == 'charter-change':
        # https://governance.openstack.org/tc/reference/charter.html#amendment
        parts = []

        # At least 2/3 positive votes.
        necessary_votes = math.ceil(TC_SIZE / 3 * 2)
        votes_to_approve = votes[1] > necessary_votes

        # Wait least 3 days after reaching majority.
        reached_supermajority = when_majority(change, necessary_votes)
        if reached_supermajority:
            earliest = str(
                reached_supermajority.date() + datetime.timedelta(4)
            )
            since_supermajority = now.date() - reached_supermajority.date()
            time_to_approve = since_supermajority > datetime.timedelta(3)
        else:
            time_to_approve = False
            earliest = '4 days after {} positive votes'.format(necessary_votes)

        if votes_to_approve and time_to_approve:
            parts.append('CAN APPROVE')

        if reached_supermajority and not time_to_approve:
            parts.append('enough required votes but too soon')

        # Even if we can approve it, if there are dissenting votes we
        # may want to continue discussion or refine the proposal.
        if votes[-1]:
            parts.append('dissenting votes')

        if votes[1] < necessary_votes:
            parts.append('not enough votes')
        elif votes[1] == necessary_votes:
            parts.append('minimum favorable votes')
        else:
            parts.append('enough votes')

        can_approve = ',\n'.join(parts)

    elif topic in (
        'goal-proposal',
        'code-change',
        'documentation-change',
        'election-results',
        'typo-fix',
    ):
        # https://governance.openstack.org/tc/reference/house-rules.html#community-wide-goal-proposals
        # https://governance.openstack.org/tc/reference/house-rules.html#code-changes
        # https://governance.openstack.org/tc/reference/house-rules.html#documentation-changes
        # https://governance.openstack.org/tc/reference/house-rules.html#election-results
        # https://governance.openstack.org/tc/reference/house-rules.html#typo-fixes

        earliest = 'Can be voted anytime'

        if votes[-1] or code_reviews[-1]:
            can_approve = 'dissenting votes'
        elif votes[1] < 2:
            can_approve = 'not enough reviews'
        else:
            can_approve = 'CAN APPROVE'

    elif topic in delegates.keys():
        # https://governance.openstack.org/tc/reference/house-rules.html#delegated-metadata
        approver_name = delegates[topic]
        can_approve = 'delegated to {}'.format(approver_name)
        if has_approved(approver_name, change):
            can_approve += '\nYES'
        elif has_rejected(approver_name, change):
            can_approve += '\nNO - delegate voted against'
        elif has_commented(approver_name, change):
            can_approve += '\ndelegate has commented'

    elif topic in ('project-update', 'new-project'):
        # https://governance.openstack.org/tc/reference/house-rules.html#other-project-team-updates

        if votes[-1] or code_reviews[-1]:
            can_approve = 'dissenting votes'
        elif votes[1] < 2:
            can_approve = 'not enough reviews'
        else:
            can_approve = 'CAN APPROVE'

    elif topic == 'goal-update':
        # https://governance.openstack.org/tc/reference/house-rules.html#goal-updates-from-ptls

        # At least 7 days old.
        earliest = str(latest_created + creation_age_threshold)

        if votes[-1]:
            can_approve = 'dissenting votes'
        elif age <= datetime.timedelta(7):
            can_approve = 'too soon'
        else:
            can_approve = 'CAN APPROVE'

    else:
        topic = 'unknown topic'
        can_approve = 'unknown topic'

    votes = '\n'.join([
        'CR:' + format_votes(code_reviews),
        ' V:' + format_votes(votes),
    ])

    tc_member_votes = {}
    for member in tc_members:
        name = member['name']
        if has_approved(member, change):
            tc_member_votes[name] = '+'
        elif has_rejected(member, change):
            tc_member_votes[name] = '-'
        elif has_commented(member, change):
            tc_member_votes[name] = 'C'
        else:
            tc_member_votes[name] = ' '
    member_votes = '\n'.join(
        '{} : {}'.format(value, name)
        for name, value in sorted(tc_member_votes.items())
    )

    return {
        'Topic': topic,
        'Subject': subject,
        'Summary': '\n'.join([
            subject.strip(),
            '',
            url,
            'Submitted by: {}'.format(owner.strip())
        ]),
        'Owner': owner,
        'URL': url,
        'Age': age.days,
        'Date': latest_created.date(),
        'Can Approve': can_approve,
        'Status': '\n'.join([topic, can_approve,
                             '{} days old'.format(age.days),
                             'earliest: {}'.format(earliest)]),
        'Earliest': earliest,
        'Votes': votes,
        'Members': member_votes,
    }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--verbose', '-v',
        action='store_const',
        dest='log_level',
        default=logging.INFO,
        const=logging.DEBUG,
    )

    args = parser.parse_args()
    logging.basicConfig(
        level=args.log_level,
    )

    tc_members = members.parse_members_file('./reference/members.yaml')
    # NOTE(mnaser): In order to consistently and properly track the votes,
    #               we need to lookup every users Gerrit account ID based
    #               on their email.
    for member in tc_members:
        raw = requests.get(
            'https://review.opendev.org/accounts/%s' % member['email'],
            headers={'Accept': 'application/json'},
        )
        member['gerritid'] = decode_json(raw).get('_account_id')

    gov = governance.Governance.from_local_repo()
    release_team = gov.get_team('Release Management')

    delegates = {
        'release-management': release_team.ptl['name'],
    }
    for tag, name in sorted(delegates.items()):
        print('Delegating {} tags to {}'.format(tag, name))

    status = sorted(
        (get_one_status(change, delegates, tc_members)
         for change in all_changes()),
        key=operator.itemgetter('URL'),
    )

    columns = (
        'Summary',
        'Status',
        'Votes',
        'Members',
    )

    x = prettytable.PrettyTable(
        field_names=columns,
        hrules=prettytable.ALL,
    )
    x.align['Summary'] = 'l'
    x.align['Status'] = 'l'
    x.align['Votes'] = 'l'
    x.align['Members'] = 'l'
    for row in status:
        x.add_row([row[c] for c in columns])
    print(x.get_string())


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A tool that will translate a projects.yaml file into a visualized graph.

Usage: universe_dot.py <projects-file> <output-file> just-project? ...

To run make sure pydot (or pydot2) is installed, as well as six and pyyaml.

After these are installed run this like:

$ python tools/universe_dot.py reference/projects.yaml \
  ironic_universe.dot ironic;

$ neato -Tsvg ironic_universe.dot > ironic_universe.svg

Then open ironic_universe.svg in your favorite web browser.

Or to create the whole openstack universe.

$ python tools/universe_dot.py reference/projects.yaml os_universe.dot

$ neato -Tsvg os_universe.dot > os_universe.svg

Then open os_universe.svg in your favorite web browser.
"""

import sys
import time

import pydot
import six
import yaml

if len(sys.argv) == 1:
    print(__doc__.strip())
    sys.exit(1)


projects_file = sys.argv[1]
output_dot_file = sys.argv[2]
just_projects = set(sys.argv[3:])

print("Reading projects from '%s'" % projects_file)
with open(projects_file, 'rb') as fh:
    projects = yaml.safe_load(fh)
    project_names = sorted(six.iterkeys(projects))
    if just_projects:
        print("Restricting to just %s projects" % list(just_projects))
        for project_name in list(project_names):
            if project_name not in just_projects:
                project_names.remove(project_name)
    project_names = tuple(project_names)

graph_nodes = {}
start = time.time()
print("Constructing nodes.")
print("Please wait...")
deliverable_node_attrs = {
    'fontsize': '11',
}
project_node_attrs = {
    'fontsize': '11',
    'shape': 'diamond',
}
node_count = 0
edges_needed = 0
for project_name in project_names:
    node = pydot.Node(name=project_name, **project_node_attrs)
    node_count += 1
    graph_nodes[project_name] = node
    deliverables = projects[project_name].get('deliverables', [])
    deliverable_names = sorted(six.iterkeys(deliverables))
    for deliverable_name in deliverable_names:
        edges_needed += 1
        node_path = "%s/%s" % (project_name, deliverable_name)
        if deliverable_name == project_name:
            # Avoid creating a self-link, since its actually not, avoid
            # this by creating a node with a hidden name, and a label that
            # is its actual name...
            node_deliverable_node_attrs = dict(deliverable_node_attrs)
            node_deliverable_node_attrs['name'] = '__%s' % deliverable_name
            node_deliverable_node_attrs['label'] = deliverable_name
        else:
            node_deliverable_node_attrs = dict(deliverable_node_attrs)
            node_deliverable_node_attrs['name'] = deliverable_name
        node = pydot.Node(**node_deliverable_node_attrs)
        graph_nodes[node_path] = node
        node_count += 1

print("Inserting %s nodes and %s edges into"
      " a new graph." % (node_count, edges_needed))
print("Please wait...")
if just_projects:
    nice_names = [name.title() for name in sorted(just_projects)]
    graph_name = '%s Universe' % (" and ".join(nice_names))
else:
    graph_name = 'OpenStack Universe'
graph_kwargs = {
    'rankdir': 'LR',
    'nodesep': '0.25',
    'overlap': 'false',
    'ranksep': '0.5',
    'splines': 'true',
    'ordering': 'in',
    'graph_name': graph_name,
}
graph = pydot.Graph(**graph_kwargs)
for project_name in project_names:
    print("  Inserting nodes for '%s'" % project_name)
    node = graph_nodes[project_name]
    graph.add_node(node)
    deliverables = projects[project_name].get('deliverables', [])
    deliverable_names = sorted(six.iterkeys(deliverables))
    for deliverable_name in deliverable_names:
        print("    Inserting node for '%s'" % deliverable_name)
        node_path = "%s/%s" % (project_name, deliverable_name)
        deliverable_node = graph_nodes[node_path]
        graph.add_node(deliverable_node)
        print("    Inserting edge for '%s' -> '%s'"
              % (project_name, deliverable_name))
        graph.add_edge(pydot.Edge(node, deliverable_node,
                                  style='dotted'))
end = time.time()
print("Finished in %0.2f seconds" % (end - start))

print("Writing graph to '%s'" % output_dot_file)
with open(output_dot_file, "w") as fh:
    fh.write(graph.to_string())
